
import logging
# import pickle
import dill as pickle
import selenium.common.exceptions
import selenium.webdriver
import sys
import time
import traceback

LOGGER = logging.getLogger('wabot')

USER_AGENT = "Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 6.0; Trident/5.0)"
REFUSE_AFTER_EXCEPTION = True
PICKLE_FILENAME = '/tmp/nhsnwebdriverdump'
EXECUTOR_PORT = 4444
REMOTE_EXECUTOR = 'http://127.0.0.1:%s/wd/hub'

class BrowserProxy:
    def __init__(self, session_name='webdriver', pickle_filename=PICKLE_FILENAME):
        self._pickle_filename = pickle_filename
        LOGGER.info('creating selenium browser: session name = %s', session_name)
        try:
            driver_type = "remote_chromium2"
            self.driver_type = driver_type
            print('session_name:', session_name)
            self.driver = self.get_driver(driver_type, session_name)
            if not self.driver:
                LOGGER.error("failed to get selenium webdriver")
                self.good = False
                return
        except Exception as ex:
            print("caught exception here")
            print(type(ex), ex)

        # self.page = nhsn_lo.pages.Login(self)
        # self.good = True

    def get_page(self):
        return type(self.page)

    def set_page(self, page):
        LOGGER.trace('switching page to %s' % page)
        newpage = page(self)
        if hasattr(newpage, 'verify'):
            rc = newpage.verify()
            if not rc:
                LOGGER.error('failed to verify page: %s' % page)
                # input('failure')
                return False
            else:
                LOGGER.trace('verified page: %s' % page)

        self.page = newpage
        return True

    def switch_after(self, fn=None, page=None):
        if not fn:
            return lambda fn: self.switch_after(fn, page = page)

        def wrapper(*args, **kwargs):
            self.set_page(page)
            return fn(*args, **kwargs)
            
        return wrapper

    def __getattr__(self, name):
        """
        Access method on current page.

        When object.method is referenced, python invokes
        object.__getattribute__ which checks the instance
        for a matching property, If none is found, python
        invokes object.__getattr__, which is overridden here.
        If the attribute is not found in the main class, this
        will search the page object for the method. This
        allows invoking page methods through the proxy class.

        nhsn.foo() invokes nhsn.page.foo()
        """
        # todo(mg) would this be a good spot to inject page verification?
        # idea: maybe we could decorate method with error checking and
        # error collection? method can return value, rc and this can
        # queue actions and error check them as it goes? At the end,
        # the user would be able to pull out the errors and check them
        if REFUSE_AFTER_EXCEPTION and not self.good:
            LOGGER.warn("broken state - refusing to invoke page action: %s" % (name))
            return

        try:
            method = getattr(self.page, name)
            return method
        except AttributeError as ex:
            raise

    def __getitem__(self, key):
        """
        Access element on the current page.
        """
        return self.page[key]

    def perform(self, meth, *args, **kwargs):
        """
        Calls self.page.method and catches exceptions.
        """
        # input('<breakpoint> at perform')
        if REFUSE_AFTER_EXCEPTION and not self.good:
            LOGGER.warn("broken state - refusing to invoke page action: %s" % (meth))
            return

        try:
            x = getattr(self.page, meth)
        except AttributeError as ex:
            LOGGER.error("Failed to invoke page action. "\
                                 "page = %s, method = %s" %\
                                 (self.page.__class__, meth))
            return None

        # Hooks:
        # x = self.page.detect_logged_out()
        # y = self.page.detected_logged_in()
        # print('logged out/in: ', x, y)

        LOGGER.trace("invoking %s.%s" % (self.page.__class__.__name__, meth))
        try:
            return x(*args, **kwargs)
        except selenium.common.exceptions.NoSuchWindowException:
            self.good = False
        except Exception as ex:
            print("caught exception: %s" % (type(ex)))
            exc_type, exc_value, exc_traceback = sys.exc_info()
            lines = traceback.format_exception(exc_type, exc_value,
                exc_traceback)
            for line in lines:
                print(line, end = "")

            self.good = False

    def reset_good_status(self):
        self.good = True

    def get_driver(self, browser, session_name='webdriver'):
        """
        Returns selenium objects of different types

        The combinations of selenium{1,2} {chrome,firefox} are supported.
        In the case of selenium 1 (selenium server), the host is a constant
        defined at the top of this file.

        Args:
            browser: The selenium webdriver requested.
                     If a chromium2 webdriver is requested, selenium
                     will try to use chromium as the browser. If
                     firefox2 is requested, selenium will try to use
                     firefox. Chromium is very fast but it is designed
                     to be fast for an interactive user which means it
                     is fast at the cost of processing power and
                     ram. Firefox is about as fast but uses less
                     resources.

                     Available browser drivers are:
                     chromium2, chromium1, firefox2, firefox1

                     The associated webdriver has to be installed
                     and runnable on the system.
        Returns:
            The selenium webdriver handle.
        """
        # nhsn actively rejects http requests that do not request with an
        # IE user agent. All of these browser instances have to change the
        # user agent in various ways.
        LOGGER.debug("creating selenium driver: %s" % (browser))
        # todo(mathew guest) turn USER_AGENT into local variable so that it
        # could be parameterized someday
        # user_agent = USER_AGENT

        # todo(mathew guest) fallback drivers if one doesn't exist?

        driver = None

        if browser == "chromium2": # Selenium 2 - Chrome
            driver = self._create_driver_chromium2()

        elif browser == 'remote_chromium2':
            driver = self._create_driver_remote_chromium2(session_name)

        elif browser == "chromium1": # Selenium 1 - Chrome without working user agent switch
            driver = self._create_driver_chromium1()

        elif browser == "firefox2": # Selenium 2 - Firefox
            driver = self._create_driver_firefox2()

        elif browser == "firefox1": # Selenium 1 - Firefox
            driver = self._create_driver_firefox1()
        else:
            LOGGER.error("an attempt was made to request an "\
                              "unsupported (by this product) selenium "\
                              "webdriver; refusing. requested = %s"\
                              % (browser))
            
        driver.implicitly_wait(10)
        return driver

    def _create_driver_remote_chromium2(self, session_name):
        print('got this far', session_name)
        fp = None
        drivers = {}
        try:
            fp = open(self._pickle_filename, 'rb')
            drivers = pickle.load(fp)
            if not drivers:
                raise Exception
            LOGGER.debug('found pickled drivers')
            driver = drivers.get(session_name)
            if not driver:
                raise Exception
            LOGGER.debug('connected to pickled webdriver instance')
            url = driver.current_url # throw error if driver isn't reliable anymore
            LOGGER.info('webdriver instance is ready')
            self.driver = driver
            return driver
        except (FileNotFoundError, IOError) as ex:
            self.driver = None
            LOGGER.error('unable to connect to existing webdriver: no pickled drivers found')
        except Exception as ex:
            self.driver = None
            LOGGER.error('unable to connect to existing webdriver: %s' % ex)

        if self.driver is None:
            print('drivers', drivers)
            LOGGER.info('creating new webdriver')
            opt = selenium.webdriver.chrome.options.Options()
            opt.add_argument("--user-agent=" + USER_AGENT)
            opt.add_argument("--kiosk-printing")
            opt.add_argument("--focus-existing-tab-on-open=false")
            driver = selenium.webdriver.Remote(
                command_executor=REMOTE_EXECUTOR % (EXECUTOR_PORT),
                desired_capabilities = opt.to_capabilities())
            print(REMOTE_EXECUTOR % EXECUTOR_PORT)
            fp = open(self._pickle_filename, 'wb')
            drivers[session_name] = driver
            print('b4 pickle')
            print(drivers)
            # print('#skipping pickle')
            pickle.dump(drivers, fp)
            # print('after pickle')
            self.driver = driver
            return driver


    def _create_driver_chromium2(self):
        opt = selenium.webdriver.chrome.options.Options()
        opt.add_argument("--user-agent=" + USER_AGENT)
        opt.add_argument("--kiosk-printing")
        driver = selenium.webdriver.Chrome(chrome_options = opt)
        self.driver = driver
        return driver

    def _create_driver_chromium1(self):
        # Selenium 1 - Chrome without working user agent switch
        # These two methods of creation ChromeOptions are equivalent objects
        options = selenium.webdriver.ChromeOptions()
        options.add_argument("--user-agent=" + USER_AGENT)
        driver = selenium.webdriver.Remote(desired_capabilities = options.to_capabilities())
        driver = selenium.webdriver.Remote(SELENIUM1_SERVER_PATH,
            selenium.webdriver.DesiredCapabilities.CHROME.copy())
        return driver

    def _create_driver_firefox2(self):
        # tmp = selenium.webdriver.FirefoxProfile()
        # tmp = None
        profile = None
        # filename = "/tmp/firefox_profile"
        # try:
            # fp = open(filename, "rb")
            # profile = pickle.load(fp)
        # except:
            # pass

        if not profile:
            profile = selenium.webdriver.FirefoxProfile(profile_directory = "/home/mathew/firefox_prof")
            profile.set_preference("general.useragent.override", USER_AGENT )
            profile.set_preference("browser.helperApps.neverAsk.saveToDisk", "text/csv");
            profile.set_preference("network.http.redirection-limit", "0" )
            # profile.set_preference("javascript.enabled", False )
            # profile.set_preference("print.always_print_silent", True)
            profile.set_preference("print.print_to_file", True)
            profile.set_preference("print.print_to_filename", "/tmp/print.pdf")
            profile.update_preferences()
            profile.set_preference("network.http.redirection-limit", "0" )
            # with open("/tmp/firefox_profile", "wb") as fp:
                # pickle.dump(profile, fp, pickle.HIGHEST_PROTOCOL)

        # driver = selenium.webdriver.Firefox()
        driver = selenium.webdriver.Firefox(profile)
        return driver

    def _create_driver_firefox1(self):
        profile = selenium.webdriver.FirefoxProfile()
        profile.set_preference("general.useragent.override", USER_AGENT )
        driver = selenium.webdriver.Remote(SELENIUM1_SERVER_PATH,
            selenium.webdriver.DesiredCapabilities.FIREFOX.copy(), browser_profile = profile )
        return driver

